part of spark_modal;

const DatePickerTheme _defaultDatePickerTheme = DatePickerTheme();

/// date picker default modal shape
const RoundedRectangleBorder _defaultPickerShape = RoundedRectangleBorder(
  borderRadius: BorderRadius.only(
    topLeft: Radius.circular(8),
    topRight: Radius.circular(8),
  ),
);

/// open date or date time picker
/// [T] is return data type, you can use [String] or [DateTime]
Future<T?> showSparkDatePicker<T>({
  required BuildContext context,
  double height = 260,
  DateTime? defaultDate,
  DateTime? maxDate,
  DateTime? minDate,
  double itemHeight = 40,
  DatePickerTheme theme = _defaultDatePickerTheme,
  DatePickerBehavior behavior = DatePickerBehavior.confirm,

  /// [SparkDateFormat] format
  String? format,
  ShapeBorder shape = _defaultPickerShape,
}) {
  assert(T == String || T == DateTime, 'date picker must setup return type');
  return showSparkModalSheet<T>(
    context: context,
    shape: shape,
    builder: (_) => SparkDatePicker<T>(
      defaultDate: defaultDate,
      height: height,
      theme: theme,
      minDate: minDate,
      maxDate: maxDate,
      itemHeight: itemHeight,
      behavior: behavior,
      format: format ?? SparkDateFormat.yearMonthDayAndHourMinuteSecond,
      onChanged: (_value) {
        Navigator.of(_).pop(_value);
      },
    ),
  );
}

class SparkDatePicker<T> extends StatefulWidget {
  const SparkDatePicker({
    Key? key,
    this.defaultDate,
    this.minDate,
    this.maxDate,
    this.height = 260,
    this.itemHeight = 40,
    required this.theme,
    required this.behavior,
    required this.format,
    this.onChanged,
    this.onChange,
  }) : super(key: key);

  /// on picker result change
  final Function(T? result)? onChanged;

  /// default date
  final DateTime? defaultDate;

  /// min date
  final DateTime? minDate;

  /// max date
  final DateTime? maxDate;

  /// picker option widget height
  final double itemHeight;

  /// widget height
  final double height;

  final DatePickerTheme theme;

  /// return behavior
  final DatePickerBehavior behavior;

  /// date format
  final String format;

  /// call is not when picker animation stop
  final Function(Iterable<SparkDatePickerData>)? onChange;

  @override
  _SparkDatePickerState<T> createState() => _SparkDatePickerState<T>();
}

class _SparkDatePickerState<T> extends State<SparkDatePicker<T>> {
  _DatePickerModel? _model;

  _DatePickerModel get model => _model!;

  final HashMap<String, Widget> _widgetsCache = HashMap();

  _DateFormatTestResult? _showResult;

  _DateFormatTestResult get showResult => _showResult!;

  @override
  initState() {
    _showResult = _DateFormatTestResult(widget.format);
    _model = _DatePickerModel(
      defaultDate: widget.defaultDate ?? DateTime.now(),
      minDate: widget.minDate,
      maxDate: widget.maxDate,
      columnData: _showResult!,
      theme: widget.theme,
    );
    super.initState();
  }

  void _onConfirm() {
    _DatePickerFormatter<T> _formatter = _DatePickerFormatter<T>(
      model._getRowData(),
      showResult,
      model.defaultDate,
    );
    if (widget.onChanged != null) {
      widget.onChanged!(_formatter.result);
    }
  }

  void _onCancel() {
    if (widget.onChanged != null) {
      widget.onChanged!(null);
    }
  }

  Future<bool> _willPop() {
    if (widget.behavior == DatePickerBehavior.confirmWithPop) {
      _onConfirm();
    }
    return Future.value(true);
  }

  Widget _buildPickerItem(
      BuildContext context, List<SparkDatePickerData> data, int index) {
    // cache key
    String _key = data[index].label + '_' + data[index].value.toString();
    if (_widgetsCache.containsKey(_key)) {
      return _widgetsCache[_key]!;
    }
    Widget _widget = Container(
      height: widget.itemHeight,
      alignment: Alignment.center,
      child: Text(
        data[index].label,
        style: widget.theme.style,
        textAlign: TextAlign.start,
      ),
    );
    // cache widget
    _widgetsCache[_key] = _widget;
    return _widget;
  }

  Widget _buildPickerColumn(_DatePickerColumn _column, int _columnIndex) {
    _DatePickerColumnNotifier? _notifier;
    switch (_column) {
      case _DatePickerColumn.year:
        _notifier = model.yearNotifier;
        break;
      case _DatePickerColumn.month:
        _notifier = model.monthNotifier;
        break;
      case _DatePickerColumn.day:
        _notifier = model.dayNotifier;
        break;
      case _DatePickerColumn.hour:
        _notifier = model.hourNotifier;
        break;
      case _DatePickerColumn.minute:
        _notifier = model.minuteNotifier;
        break;
      case _DatePickerColumn.second:
        _notifier = model.secondNotifier;
        break;
    }
    return Expanded(
      flex: 1,
      child: ValueListenableBuilder(
        valueListenable: _notifier,
        builder: (_ctx, _value, _) {
          return SparkPicker(
            key: ValueKey(_notifier!.updateKey),
            controller: _notifier.controller,
            itemHeight: widget.itemHeight,
            itemBuilder: (_c, _index) {
              return _buildPickerItem(_c, _notifier!.list, _index);
            },
            itemCount: _notifier.list.length,
            onStart: () {
              _notifier!._startAnimation();
              model._requestUpdate(_column, () {
                model.handleColumnChanged(_columnIndex);
              });
            },
            onChange: (_) => _notifier?.justUpdateValue(_),
            onChanged: (_) {
              _notifier!._onAnimationEnd();
              model._onAnimationEnd(_column);
            },
          );
        },
      ),
    );
  }

  Widget _buildTopBar(BuildContext context) {
    return Container(
      padding: widget.theme.topBarPadding,
      child: Row(
        mainAxisAlignment: MainAxisAlignment.spaceBetween,
        children: [
          GestureDetector(
            behavior: HitTestBehavior.opaque,
            onTap: () {
              _onCancel();
            },
            child: Container(
              decoration: widget.theme.cancelDecoration,
              child: Text(
                widget.theme.cancelText,
                style: widget.theme.cancelStyle,
              ),
            ),
          ),
          GestureDetector(
            behavior: HitTestBehavior.opaque,
            onTap: () {
              _onConfirm();
            },
            child: Container(
              decoration: widget.theme.confirmDecoration,
              child: Text(
                widget.theme.confirmText,
                style: widget.theme.confirmStyle,
              ),
            ),
          )
        ],
      ),
    );
  }

  @override
  Widget build(BuildContext context) {
    return WillPopScope(
      child: SizedBox(
        height: widget.height,
        child: Column(
          children: [
            _buildTopBar(context),
            Expanded(
              child: Row(
                children: [
                  for (int _columnIndex = 0;
                      _columnIndex < showResult.columnList.length;
                      _columnIndex++)
                    _buildPickerColumn(
                        showResult.columnList[_columnIndex], _columnIndex),
                ],
              ),
              flex: 1,
            ),
          ],
        ),
      ),
      onWillPop: _willPop,
    );
  }
}

class _DatePickerModel {
  _DatePickerModel({
    required this.defaultDate,
    this.minDate,
    this.maxDate,
    required this.theme,
    required this.columnData,
  }) {
    _initDatePickerData();
  }

  final DatePickerTheme theme;

  final _DateFormatTestResult columnData;

  List<_DatePickerColumn> get showColumn => columnData.columnList;

  final DateTime defaultDate;

  final DateTime? minDate;

  final DateTime? maxDate;

  final List<DateTime> _dateRange = [];

  DateTime get getStartDate => _dateRange.first;

  DateTime get getEndDate => _dateRange.last;

  final List<int> _maxMount = [1, 3, 5, 7, 8, 10, 12];

  bool get _inAnimation =>
      yearNotifier._inAnimation ||
      monthNotifier._inAnimation ||
      dayNotifier._inAnimation ||
      minuteNotifier._inAnimation ||
      hourNotifier._inAnimation ||
      secondNotifier._inAnimation;

  final _UpdateTaskManage _beforeTask = _UpdateTaskManage();

  _DatePickerColumnNotifier yearNotifier = _DatePickerColumnNotifier(null),
      monthNotifier = _DatePickerColumnNotifier(null),
      dayNotifier = _DatePickerColumnNotifier(null),
      hourNotifier = _DatePickerColumnNotifier(null),
      minuteNotifier = _DatePickerColumnNotifier(null),
      secondNotifier = _DatePickerColumnNotifier(null);

  void _initDatePickerData() {
    if (minDate != null) {
      _dateRange.add(minDate!);
    }
    if (maxDate != null) {
      if (_dateRange.isEmpty) {
        _dateRange.add(DateTime(
          defaultDate.year - 50,
          defaultDate.month,
          defaultDate.day,
        ));
      }
      _dateRange.add(maxDate!);
    }
    if (_dateRange.length == 1) {
      _dateRange.add(DateTime(
        defaultDate.year + 30,
        defaultDate.month,
        defaultDate.day,
        23,
        59,
        59,
      ));
    }
    if (_dateRange.isEmpty) {
      _dateRange.addAll(getDateRange(20, 30));
    }
    DateTime _start = _dateRange.first;
    DateTime _end = _dateRange.last;
    yearNotifier.list = listFillByNumberRange(
      _start.year,
      _end.year,
      suffix: theme.yearSuffix,
    );
    _setUpDefaultDate();
    _changeNextColumnData(0);
    _getDefaultIndex();
  }

  void _setUpDefaultDate() {
    yearNotifier.value =
        SparkDatePickerData.form('${defaultDate.year}', defaultDate.year, 0);
    monthNotifier.value =
        SparkDatePickerData.form('${defaultDate.month}', defaultDate.month, 0);
    dayNotifier.value =
        SparkDatePickerData.form('${defaultDate.day}', defaultDate.day, 0);
    hourNotifier.value =
        SparkDatePickerData.form('${defaultDate.hour}', defaultDate.hour, 0);
    minuteNotifier.value = SparkDatePickerData.form(
        '${defaultDate.minute}', defaultDate.minute, 0);
    secondNotifier.value = SparkDatePickerData.form(
        '${defaultDate.second}', defaultDate.second, 0);
  }

  void _getDefaultIndex() {
    if (columnData.hasYear) {
      yearNotifier._getDefaultByDate(defaultDate.year);
    }
    if (columnData.hasMonth) {
      monthNotifier._getDefaultByDate(defaultDate.month);
    }
    if (columnData.hasDay) {
      dayNotifier._getDefaultByDate(defaultDate.day);
    }
    if (columnData.hasHour) {
      hourNotifier._getDefaultByDate(defaultDate.hour);
    }
    if (columnData.hasMinute) {
      minuteNotifier._getDefaultByDate(defaultDate.minute);
    }
    if (columnData.hasSecond) {
      secondNotifier._getDefaultByDate(defaultDate.second);
    }
  }

  void getMonthListByYear() {
    int year = yearNotifier.targetValue;
    int _start = 1;
    int _end = 12;
    if (year == getStartDate.year) {
      _start = getStartDate.month;
    } else if (year == getEndDate.year) {
      _end = getEndDate.month;
    }
    if (monthNotifier.needUpdateList(_start, _end)) {
      monthNotifier.list = listFillByNumberRange(
        _start,
        _end,
        suffix: theme.monthSuffix,
      );
    }
  }

  void getDayByYearAndMonth() {
    int year = yearNotifier.targetValue;
    int month = monthNotifier.targetValue;
    bool _isLeapYear = isLeapYear(year);
    int _start = 1;
    int _end = 31;
    bool _isStartYearAndMonth =
        getStartDate.year == year && getStartDate.month == month;
    bool _isEndYearAndMonth =
        getEndDate.year == year && getEndDate.month == month;
    if (_isStartYearAndMonth) {
      _start = getStartDate.day;
    }
    if (_isEndYearAndMonth) {
      _end = getEndDate.day;
    }
    if (month == 2) {
      if (_isStartYearAndMonth || !_isEndYearAndMonth) {
        _end = _isLeapYear ? 29 : 28;
      }
    } else if (_maxMount.contains(month)) {
      if (_isStartYearAndMonth || !_isEndYearAndMonth) {
        _end = 31;
      }
    } else if (_isStartYearAndMonth || !_isEndYearAndMonth) {
      _end = 30;
    }
    if (dayNotifier.needUpdateList(_start, _end)) {
      dayNotifier.list = listFillByNumberRange(
        _start,
        _end,
        suffix: theme.daySuffix,
      );
    }
  }

  void getHourByDate() {
    int year = yearNotifier.targetValue;
    int month = monthNotifier.targetValue;
    int day = dayNotifier.targetValue;
    bool _isStartDate = getStartDate.year == year &&
        getStartDate.month == month &&
        getStartDate.day == day;
    bool _isEndDate = getEndDate.year == year &&
        getEndDate.month == month &&
        getEndDate.day == day;
    int _start = 0;
    int _end = 23;
    if (_isStartDate) {
      _start = getStartDate.hour;
    }
    if (_isEndDate) {
      _end = getEndDate.hour;
    }
    if (hourNotifier.needUpdateList(_start, _end)) {
      hourNotifier.list = listFillByNumberRange(_start, _end);
    }
  }

  void getMinuteByHourAndDate() {
    int year = yearNotifier.targetValue;
    int month = monthNotifier.targetValue;
    int day = dayNotifier.targetValue;
    int hour = hourNotifier.targetValue;
    bool _isStartDate = getStartDate.year == year &&
        getStartDate.month == month &&
        getStartDate.day == day &&
        getStartDate.hour == hour;
    bool _isEndDate = getEndDate.year == year &&
        getEndDate.month == month &&
        getEndDate.day == day &&
        getEndDate.hour == hour;
    int _start = 0;
    int _end = 59;
    if (_isStartDate) {
      _start = getStartDate.minute;
    }
    if (_isEndDate) {
      _end = getEndDate.minute;
    }
    if (minuteNotifier.needUpdateList(_start, _end)) {
      minuteNotifier.list = listFillByNumberRange(_start, _end);
    }
  }

  void getSecondByHourMinuteAndDate() {
    int year = yearNotifier.targetValue;
    int month = monthNotifier.targetValue;
    int day = dayNotifier.targetValue;
    int hour = hourNotifier.targetValue;
    int minute = minuteNotifier.targetValue;
    bool _isStartDate = getStartDate.year == year &&
        getStartDate.month == month &&
        getStartDate.day == day &&
        getStartDate.hour == hour &&
        getStartDate.minute == minute;
    bool _isEndDate = getEndDate.year == year &&
        getEndDate.month == month &&
        getEndDate.day == day &&
        getEndDate.hour == hour &&
        getEndDate.minute == minute;
    int _start = 0;
    int _end = 59;
    if (_isStartDate) {
      _start = getStartDate.second;
    }
    if (_isEndDate) {
      _end = getEndDate.second;
    }
    if (secondNotifier.needUpdateList(_start, _end)) {
      secondNotifier.list = listFillByNumberRange(_start, _end);
    }
  }

  void _changeNextColumnData(int startIndex) {
    List<_DatePickerColumn> _columnData = columnData.columnList;
    if (_columnData.length <= startIndex) {
      return;
    }
    _DatePickerColumn _columnName = _columnData[startIndex];
    switch (_columnName) {
      case _DatePickerColumn.year:
        getMonthListByYear();
        _changeNextColumnData(startIndex + 1);
        break;
      case _DatePickerColumn.month:
        getDayByYearAndMonth();
        _changeNextColumnData(startIndex + 1);
        break;
      case _DatePickerColumn.day:
        getHourByDate();
        _changeNextColumnData(startIndex + 1);
        break;
      case _DatePickerColumn.hour:
        getMinuteByHourAndDate();
        _changeNextColumnData(startIndex + 1);
        break;
      case _DatePickerColumn.minute:
        getSecondByHourMinuteAndDate();
        _changeNextColumnData(startIndex + 1);
        break;
      case _DatePickerColumn.second:
        break;
    }
  }

  bool isLeapYear(int year) {
    return (year % 4 == 0 && year % 100 != 0) || year % 400 == 0;
  }

  void handleColumnChanged(int index) {
    _changeNextColumnData(index);
  }

  void _requestUpdate(_DatePickerColumn type, Function task) {
    if (_inAnimation) {
      switch (type) {
        case _DatePickerColumn.year:
          _beforeTask._yearTask = task;
          break;
        case _DatePickerColumn.month:
          _beforeTask._monthTask = task;
          break;
        case _DatePickerColumn.day:
          _beforeTask._dayTask = task;
          break;
        case _DatePickerColumn.hour:
          _beforeTask._hourTask = task;
          break;
        case _DatePickerColumn.minute:
          _beforeTask._minuteTask = task;
          break;
        case _DatePickerColumn.second:
          _beforeTask._secondTask = task;
          break;
      }
    } else {
      task();
    }
  }

  void _onAnimationEnd(index) {
    if (!_inAnimation) {
      _beforeTask.runTask();
    }
  }

  Iterable<SparkDatePickerData> _getRowData() {
    List<SparkDatePickerData> _list = [];
    for (_DatePickerColumn _column in showColumn) {
      switch (_column) {
        case _DatePickerColumn.year:
          _list.add(yearNotifier.value);
          break;
        case _DatePickerColumn.month:
          _list.add(monthNotifier.value);
          break;
        case _DatePickerColumn.day:
          _list.add(dayNotifier.value);
          break;
        case _DatePickerColumn.hour:
          _list.add(hourNotifier.value);
          break;
        case _DatePickerColumn.minute:
          _list.add(minuteNotifier.value);
          break;
        case _DatePickerColumn.second:
          _list.add(secondNotifier.value);
          break;
      }
    }
    return _list;
  }
}

/// generate string list by range
List<SparkDatePickerData> listFillByNumberRange(int start, int end,
    {String suffix = ''}) {
  return List.generate(
    end - start + 1,
    (index) => SparkDatePickerData.form(
      '${_digits(index + start, 2)}$suffix',
      index + start,
      index,
    ),
  );
}

/// get string format for min length
/// [length] = 2
/// [0] => '00'
/// [00] => '00'
/// [000] => '000'
String _digits(int value, int length) {
  return '$value'.padLeft(length, "0");
}

class DatePickerTheme {
  const DatePickerTheme({
    this.yearSuffix = '年',
    this.monthSuffix = '月',
    this.daySuffix = '日',
    this.style = const TextStyle(fontSize: 14, color: Color(0xFF333333)),
    this.confirmStyle = const TextStyle(fontSize: 14, color: Color(0xFF2878FF)),
    this.cancelStyle = const TextStyle(fontSize: 14, color: Color(0xFF333333)),
    this.cancelDecoration,
    this.confirmDecoration,
    this.topBarPadding =
        const EdgeInsets.symmetric(horizontal: 14, vertical: 10),
  });

  /// year picker option suffix text
  /// [year][monthSuffix]
  final String yearSuffix;

  /// month picker option suffix text
  /// [month][monthSuffix]
  final String monthSuffix;

  /// day picker option suffix text
  /// [day][daySuffix]
  final String daySuffix;

  /// picker option style
  final TextStyle style;

  /// confirm text
  final String confirmText = '确认';

  /// cancel text
  final String cancelText = '取消';

  /// confirm text text style
  final TextStyle confirmStyle;

  /// cancel text style
  final TextStyle cancelStyle;

  /// confirm container decoration
  final BoxDecoration? confirmDecoration;

  /// cancel container decoration
  final BoxDecoration? cancelDecoration;

  /// top bar container padding
  final EdgeInsets topBarPadding;
}

/// spark date picker behavior
enum DatePickerBehavior {
  /// tap [confirm] and modal pop return picker date value
  confirmWithPop,

  /// just tap [confirm] return date picker value
  confirm,
}

class SparkDateFormat {
  /// use part:
  /// 'yyyy' year
  /// 'MM' month
  /// 'dd' day
  /// 'hh' hour
  /// 'mm' minute
  /// 'ss' second
  SparkDateFormat._();

  static const String _yearPart = 'yyyy';

  static const String _monthPart = 'MM';

  static const String _dayPart = 'dd';

  static const String _hourPart = 'hh';

  static const String _minutePart = 'mm';

  static const String _secondPart = 'ss';

  /// date format as '2020-01'
  static const String yearMonth = 'yyyy-MM';

  /// date format as '2020/01'
  static const String yearMonthObliqueLine = 'yyyy/MM';

  /// date format as '2020-01-01'
  static const String yearMonthDay = 'yyyy-MM-dd';

  /// date format as '2020/01/01'
  static const String yearMonthDayObliqueLine = 'yyyy/MM/dd';

  /// time format as '12:01'
  static const String hourMinute = 'hh:mm';

  /// time format as '12:01:59'
  static const String hourMinuteSecond = 'hh:mm:ss';

  /// date time format as '2021-12-01 12:01'
  static const String yearMonthDayAndHourMinute = 'yyyy-MM-dd hh:mm';

  /// date time format as '2021/12/01 12:01'
  static const String yearMonthDayObliqueLineAndHourMinute = 'yyyy/MM/dd hh:mm';

  /// date time format as '2021-12-01 12:01:59'
  static const String yearMonthDayAndHourMinuteSecond = 'yyyy-MM-dd hh:mm:ss';

  /// date time format as '2021/12/01 12:01:59'
  static const String yearMonthDayObliqueLineAndHourMinuteSecond =
      'yyyy/MM/dd hh:mm:ss';
}

class _DatePickerFormatter<T> {
  _DatePickerFormatter(this.rawData, this.dateFormat, this.defaultDate) {
    _formatter();
  }

  DateTime defaultDate;

  Iterable<SparkDatePickerData> rawData;

  String stringResult = '';

  DateTime dateTimeResult = DateTime.now();

  _DateFormatTestResult dateFormat;

  void _formatter() {
    if (T == String) {
      stringResult =
          _buildDateStringByFormatList(dateFormat.columnList, rawData);
    }
    if (T == DateTime) {
      dateTimeResult = _buildDateTimeByData(dateFormat.columnList, rawData);
    }
  }

  /// change date picker value to [DateTime]
  DateTime _buildDateTimeByData(
      List<_DatePickerColumn> columns, Iterable<SparkDatePickerData> rawData) {
    int year = defaultDate.year,
        month = defaultDate.month,
        day = defaultDate.day,
        hour = defaultDate.hour,
        minute = defaultDate.minute,
        second = defaultDate.second;
    for (int _index = 0; _index < columns.length; _index++) {
      _DatePickerColumn _columnName = columns[_index];
      switch (_columnName) {
        case _DatePickerColumn.year:
          year = rawData.elementAt(_index).value;
          break;
        case _DatePickerColumn.month:
          month = rawData.elementAt(_index).value;
          break;
        case _DatePickerColumn.day:
          day = rawData.elementAt(_index).value;
          break;
        case _DatePickerColumn.hour:
          hour = rawData.elementAt(_index).value;
          break;
        case _DatePickerColumn.minute:
          minute = rawData.elementAt(_index).value;
          break;
        case _DatePickerColumn.second:
          second = rawData.elementAt(_index).value;
          break;
      }
    }
    return DateTime(year, month, day, hour, minute, second);
  }

  /// get column and pick data
  /// and build date string by column
  String _buildDateStringByFormatList(
      List<_DatePickerColumn> columns, Iterable<SparkDatePickerData> rawData) {
    String _tempDate = dateFormat.format;
    for (int _index = 0; _index < columns.length; _index++) {
      _DatePickerColumn _columnName = columns[_index];
      String _replace = '';
      switch (_columnName) {
        case _DatePickerColumn.year:
          _replace = SparkDateFormat._yearPart;
          break;
        case _DatePickerColumn.month:
          _replace = SparkDateFormat._monthPart;
          break;
        case _DatePickerColumn.day:
          _replace = SparkDateFormat._dayPart;
          break;
        case _DatePickerColumn.hour:
          _replace = SparkDateFormat._hourPart;
          break;
        case _DatePickerColumn.minute:
          _replace = SparkDateFormat._minutePart;
          break;
        case _DatePickerColumn.second:
          _replace = SparkDateFormat._secondPart;
          break;
      }
      _tempDate = _tempDate.replaceAll(
        _replace,
        _digits(rawData.elementAt(_index).value, 2),
      );
    }
    return _tempDate;
  }

  /// get result by [T]
  /// [T] is [String] or [DateTime]
  /// you can see [showSparkDatePicker]
  T? get result => T == String
      ? stringResult as T
      : T == DateTime
          ? dateTimeResult as T
          : null;
}

class _DateFormatTestResult {
  /// date picker format test
  /// get show column
  _DateFormatTestResult(this.format) {
    _tryTest(format);
  }

  String format;

  final List<_DatePickerColumn> _columnList = [];

  List<_DatePickerColumn> get columnList => _columnList;

  /// has year pick column
  bool hasYear = false;

  /// has month pick column
  bool hasMonth = false;

  /// has day pick column
  bool hasDay = false;

  /// has hour pick column
  bool hasHour = false;

  /// has minute pick column
  bool hasMinute = false;

  /// has second pick column
  bool hasSecond = false;

  void _tryTest(String format) {
    List<String> _list = format.split(RegExp(r'\/|\-|\:|\s'));
    for (int _index = 0; _index < _list.length; _index++) {
      String _part = _list[_index];
      switch (_part) {
        case SparkDateFormat._yearPart:
          _columnList.add(_DatePickerColumn.year);
          hasYear = true;
          break;
        case SparkDateFormat._monthPart:
          _columnList.add(_DatePickerColumn.month);
          hasMonth = true;
          break;
        case SparkDateFormat._dayPart:
          _columnList.add(_DatePickerColumn.day);
          hasDay = true;
          break;
        case SparkDateFormat._hourPart:
          _columnList.add(_DatePickerColumn.hour);
          hasHour = true;
          break;
        case SparkDateFormat._minutePart:
          _columnList.add(_DatePickerColumn.minute);
          hasMinute = true;
          break;
        case SparkDateFormat._secondPart:
          _columnList.add(_DatePickerColumn.second);
          hasSecond = true;
          break;
        default:
          break;
      }
    }
  }
}

/// date picker column type
enum _DatePickerColumn {
  year,
  month,
  day,
  hour,
  minute,
  second,
}

class _DatePickerColumnNotifier extends ChangeNotifier
    implements ValueListenable<SparkDatePickerData> {
  _DatePickerColumnNotifier(this._value);

  SparkDatePickerData? _value;

  FixedExtentScrollController controller =
      FixedExtentScrollController(initialItem: 0);

  int lastStartIndex = -1;

  int lastEndIndex = -1;

  @override
  SparkDatePickerData get value => _value!;

  set value(SparkDatePickerData newValue) {
    if (_value == newValue) {
      return;
    }
    _value = newValue;
    _notifyUpdate();
  }

  bool _inAnimation = false;

  bool needUpdateList(int start, int end) {
    bool _rangeChanged = start != lastStartIndex || end != lastEndIndex;
    if (_rangeChanged) {
      lastStartIndex = start;
      lastEndIndex = end;
    }
    return _rangeChanged;
  }

  void justUpdateValue(int index) {
    _value = list[index];
  }

  void _startAnimation() {
    _inAnimation = true;
  }

  void _onAnimationEnd() {
    _inAnimation = false;
  }

  void justUpdate() {
    notifyListeners();
  }

  int get targetValue => value.value;

  int get targetIndex => value.index;

  List<SparkDatePickerData> get list => _list;

  String updateKey = '';

  List<SparkDatePickerData> _list = [];

  set list(List<SparkDatePickerData> newValue) {
    if (newValue.isNotEmpty && newValue.length == _list.length) {
      if (newValue.first == _list.first && newValue.last == _list.last) {
        return;
      }
    }
    if (newValue.isEmpty && _list.isEmpty) {
      return;
    }
    // picker update key
    // picker date update tag
    updateKey =
        '${newValue.first.label}_${newValue.first.index}_${newValue.last.label}_${newValue.last.index}';
    _list = newValue;
    _notifyUpdate();
  }

  void _notifyUpdate() {
    _updateWhenOutOfRange();
    controller = FixedExtentScrollController(initialItem: _value?.index ?? 0);
    notifyListeners();
  }

  /// setup default value
  void _getDefaultByDate(int _value) {
    if (list.isEmpty) {
      return;
    }
    Iterable<SparkDatePickerData> _findByValue =
        list.where((element) => element.value == _value);
    if (_findByValue.isNotEmpty) {
      value = _findByValue.first;
      return;
    }
    if (value.value < _list.first.value) {
      value = _list.first;
      return;
    }
    value = _list.last;
  }

  /// index check
  /// get boundary value
  void _updateWhenOutOfRange() {
    // update check is not empty
    // setup is empty
    if (_list.isEmpty || _value == null) {
      return;
    }
    if (!_list.contains(value)) {
      Iterable<SparkDatePickerData> _findByValue =
          list.where((element) => element.value == value.value);
      if (_findByValue.isNotEmpty) {
        _value = _findByValue.first;
        return;
      }
      if (value.value < _list.first.value) {
        _value = _list.first;
        return;
      }
      _value = _list.last;
    }
  }
}

class _UpdateTaskManage {
  Function? _yearTask;
  Function? _monthTask;
  Function? _dayTask;
  Function? _hourTask;
  Function? _minuteTask;
  Function? _secondTask;

  void runTask() {
    List<Function?> _taskList = [
      _yearTask,
      _monthTask,
      _dayTask,
      _hourTask,
      _minuteTask,
      _secondTask
    ];
    var task =
        _taskList.firstWhere((element) => element != null, orElse: () => null);
    if (task != null) {
      task();
      _clearTask();
    }
  }

  void _clearTask() {
    _yearTask = null;
    _monthTask = null;
    _dayTask = null;
    _hourTask = null;
    _minuteTask = null;
    _secondTask = null;
  }
}
