part of spark;

/// create default theme data
SparkTheme _defaultThemeData = const SparkTheme();

class SparkForm extends StatefulWidget {
  /// form base widget, manage data and verify the data
  /// if SparkForm not wrap in SparkThemeData
  /// will be create SparkThemeData root
  const SparkForm({
    Key? key,
    required this.child,
    required this.manage,
    this.controller,
  }) : super(key: key);

  /// form controller
  final SparkFormController? controller;

  /// form item builder
  final Widget child;

  /// form data manage
  final FormGroupManage manage;

  @override
  State<StatefulWidget> createState() => _SparkFormBase();
}

class _SparkFormBase extends State<SparkForm> {
  @override
  void initState() {
    widget.controller?._setFormItemData(_formItemData, widget.manage);
    super.initState();
  }

  /// handle form data change
  void _handleFormDataChange(FormDataChangeNotification notification) {
    String _field = notification.changeField!;
    ValidateResult? _result = widget.controller?.validateField(
      _field,
      VerifyType.change,
      preValue: notification.newValue,
    );
    FormGroupItem _manage = widget.manage[_field];
    if (_result != null) {
      if (!_result.success) {
        if (_result.strategy == FormStrategy.limit) {
          // reset old value
          _manage.updateValue(_manage.oldValue, true);
          return;
        }
      }
    }
    _manage.updateValue(notification.newValue, true);
  }

  /// form item provider
  final SparkFormItemData _formItemData = SparkFormItemData();

  @override
  Widget build(BuildContext context) {
    bool _hasThemeData = SparkThemeData.of(context) != null;
    Widget _widget = NotificationListener<FormDataChangeNotification>(
      child: SparkFormData(
        child: widget.child,
        formData: _formItemData.._setManage(widget.manage),
      ),
      onNotification: (notification) {
        _handleFormDataChange(notification);
        return true;
      },
    );
    if (!_hasThemeData) {
      return SparkThemeData(
        theme: _defaultThemeData,
        child: _widget,
      );
    }
    return _widget;
  }

  @override
  void dispose() {
    _formItemData.dispose();
    super.dispose();
  }
}

class SparkFormItem extends StatefulWidget {
  const SparkFormItem({
    Key? key,
    required this.fieldName,
    required this.child,
    this.direction,
    this.labelBuilder,
    this.label,
    this.isRequired,
    this.rules,
    this.labelWidth,
    this.labelPadding,
    this.decoration,
    this.readOnly = false,
    this.padding,
    this.margin,
    this.labelStyle,
    this.crossCenter,
    this.errorStyle,
    this.errorContentPadding,
    this.labelSuffix,
    this.appendWidget,
    this.appendWidgetBuilder,
  }) : super(key: key);

  /// form item bind field name
  final String fieldName;

  /// form label name
  final String? label;

  /// form label widget builder
  final WidgetThemeBuilder? labelBuilder;

  /// form widget builder
  final Widget child;

  /// form layout direction, set title layout in top or left
  final Axis? direction;

  /// form item is required
  final bool? isRequired;

  /// label width
  final double? labelWidth;

  /// setup label layout padding
  final EdgeInsets? labelPadding;

  /// form item decoration
  final FocusDecoration? decoration;

  /// value check rules
  final List<SparkValidator>? rules;

  /// set form item response focusNode
  /// if false, focus style is not change
  /// [FocusPointerScope.shadow]
  final bool? readOnly;

  /// form item content padding
  final EdgeInsets? padding;

  /// form item content margin
  final EdgeInsets? margin;

  /// set label text style
  final TextStyle? labelStyle;

  /// cross use [CrossAxisAlignment.center]
  final bool? crossCenter;

  /// error tip message style
  final TextStyle? errorStyle;

  /// tip message content layout
  final EdgeInsets? errorContentPadding;

  /// label suffix, is left label text suffix
  /// not form item right suffix
  final Widget? labelSuffix;

  final Widget? appendWidget;

  final AppendChildBuilder? appendWidgetBuilder;

  @override
  State<StatefulWidget> createState() => _SparkFormItemBase();
}

class _SparkFormItemBase extends State<SparkFormItem> {
  SparkTheme? _themeData;

  bool _init = false;

  SparkTheme get _theme => _themeData!;

  SparkFormItemData? _item;

  bool _showErrorTip = false;

  String _errorMessage = '';

  void _initData() {
    if (_init) {
      return;
    }
    _init = true;
    _item = SparkFormData.of(context);
    assert(_item != null && _themeData != null,
        'form item must be wrap in the SparkForm');
    _item!.validateList[widget.fieldName] = (value, type) {
      return _validateFormItemValue(type, value);
    };
  }

  ///  run click broadcast
  void _handleFormItemClick() {
    FormGroupItem? _fieldManage = _item?.manage[widget.fieldName];
    if (_fieldManage != null && _fieldManage.clickCallback != null) {
      _fieldManage.clickCallback!();
    }
  }

  /// verify value on blur
  void _handleUnFocus() {
    FormGroupItem? _fieldManage = _item?.manage[widget.fieldName];
    _validateFormItemValue(VerifyType.blur, _fieldManage!.notifier.value);
  }

  @override
  void dispose() {
    _item?.disposeField(widget.fieldName);
    super.dispose();
  }

  /// validate value
  ValidateResult _validateFormItemValue(VerifyType type, dynamic value) {
    Iterable<SparkValidator> _rules = (type == VerifyType.all
            ? widget.rules
            : widget.rules?.where((element) =>
                element.type == type || element.type == VerifyType.all)) ??
        [];
    for (SparkValidator _rule in _rules) {
      ValidateResult _result = _rule.validate(
        value,
        widget.fieldName,
        _rule.type,
      );
      if (!_result.success) {
        if (_result.strategy == FormStrategy.tip) {
          // show error message
          setState(() {
            _showErrorTip = true;
            _errorMessage = _result.errorMessage ?? '';
          });
        }
        return _result;
      }
    }
    if (_showErrorTip) {
      setState(() {
        _showErrorTip = false;
        _errorMessage = '';
      });
    }
    return ValidateResult.from(success: true, fieldName: widget.fieldName);
  }

  /// item widget data change
  /// should notify form widget handle change
  /// form item just register validate function
  void _handleItemDataChange(FormDataChangeNotification notification) {
    FormDataChangeNotification(
      changeField: widget.fieldName,
      newValue: notification.newValue,
    ).dispatch(context);
  }

  /// render label title
  Widget? _buildLabel() {
    TextStyle _labelTextStyle = widget.labelStyle ??
        TextStyle(
          fontSize: _theme.labelFontSize,
          color: _theme.labelColor,
          fontWeight: _theme.labelFontWeight,
        );
    // build label widget
    Widget? _label;
    if (widget.label != null) {
      _label = Text(
        widget.label!,
        style: _labelTextStyle,
        overflow: TextOverflow.ellipsis,
      );
    }
    _label ??= widget.labelBuilder != null
        ? widget.labelBuilder!(context, _theme)
        : null;
    // do not build any widget
    if (_label == null && widget.isRequired != true) {
      return null;
    }
    return Row(
      mainAxisAlignment: MainAxisAlignment.start,
      crossAxisAlignment: CrossAxisAlignment.baseline,
      textBaseline: TextBaseline.alphabetic,
      children: [
        if (widget.isRequired == true)
          Text(
            _theme.requiredPrefix ?? '*',
            style: _labelTextStyle.copyWith(
              color: _theme.requiredColor,
            ),
          ),
        if (_label != null) _label,
        if (widget.labelSuffix != null) widget.labelSuffix!,
      ],
    );
  }

  Widget? _buildAppendWidget(BuildContext context, dynamic value) {
    if (widget.appendWidgetBuilder != null) {
      return widget.appendWidgetBuilder!(
        context,
        value,
        widget.appendWidget,
      );
    }
    return widget.appendWidget;
  }

  Widget _buildItemWidget(BuildContext context, BoxConstraints constraints) {
    TextStyle _errorTipStyle =
        widget.errorStyle ?? const TextStyle(fontSize: 12, color: Colors.red);
    // build label
    Widget? _label = _buildLabel();
    Widget _errorTipWidget = LayoutBuilder(
      builder: (_, _b) => AnimatedContainer(
        constraints: _b,
        padding: widget.errorContentPadding,
        height: _showErrorTip
            ? (_errorTipStyle.fontSize ?? 16) * (_errorTipStyle.height ?? 1) + 3
            : 0,
        alignment: Alignment.centerLeft,
        duration: const Duration(milliseconds: 200),
        child: Text(
          _errorMessage,
          overflow: TextOverflow.ellipsis,
          style: _errorTipStyle,
        ),
      ),
    );
    double _labelWidth = widget.labelWidth ?? _theme.labelWidth ?? 0;
    EdgeInsets? _labelPadding = widget.labelPadding ?? _theme.labelPadding;
    Widget? _directionWidget;
    // layout direction
    if (widget.direction == Axis.vertical) {
      _directionWidget = Column(
        crossAxisAlignment: CrossAxisAlignment.start,
        children: [
          if (_label != null)
            Container(
              padding: _labelPadding,
              width: _labelWidth,
              child: _label,
            ),
          widget.child,
          ValueListenableBuilder(
            valueListenable: _item!.manage[widget.fieldName].notifier,
            builder: (_ctx, _value, _) {
              Widget? _appendWidget = _buildAppendWidget(context, _value);
              return _appendWidget ?? const SizedBox();
            },
          ),
          _errorTipWidget,
        ],
      );
    }

    bool _isCrossCenter = widget.crossCenter ?? true;
    _directionWidget ??= Column(
      crossAxisAlignment: CrossAxisAlignment.start,
      children: [
        Row(
          mainAxisAlignment: MainAxisAlignment.start,
          crossAxisAlignment: _isCrossCenter
              ? CrossAxisAlignment.center
              : CrossAxisAlignment.start,
          textBaseline: TextBaseline.ideographic,
          mainAxisSize: MainAxisSize.max,
          children: [
            if (_label != null)
              Container(
                padding: _labelPadding,
                width: _labelWidth,
                child: _label,
              ),
            Expanded(
              flex: 1,
              child: widget.child,
            )
          ],
        ),
        ValueListenableBuilder(
          valueListenable: _item!.manage[widget.fieldName].notifier,
          builder: (_ctx, _value, _) {
            Widget? _appendWidget = _buildAppendWidget(context, _value);
            return _appendWidget ?? const SizedBox();
          },
        ),
        _errorTipWidget,
      ],
    );
    return _directionWidget;
  }

  @override
  Widget build(BuildContext context) {
    _themeData = SparkThemeData.of(context);
    _initData();
    assert(
      !(widget.label != null && widget.labelBuilder != null),
      'form field label and labelBuilder can only choose one',
    );

    FocusDecoration? _decoration =
        widget.decoration ?? _theme.formItemDecoration;
    if (_showErrorTip) {
      _decoration = _decoration?.copyWith(
        focusBorder: Border(
          bottom: _decoration.focusBorder!.bottom.copyWith(
            color: _theme.errorColor,
          ),
        ),
        unFocusBorder: Border(
          bottom: _decoration.focusBorder!.bottom.copyWith(
            color: _theme.errorColor,
          ),
        ),
      );
    }
    return LayoutBuilder(
      builder: (_, _box) => FocusPointerScope(
        onTap: () {
          _handleFormItemClick();
        },
        constraints: _box,
        shadow: widget.readOnly,
        margin: widget.margin ?? _theme.formItemMargin,
        padding: widget.padding ?? _theme.formItemPadding,
        decoration: _decoration,
        focusNode: _item?.manage[widget.fieldName].focusNode,
        onUnFocus: () {
          _handleUnFocus();
        },
        child: NotificationListener<FormDataChangeNotification>(
          child: _buildItemWidget(context, _box),
          onNotification: (notification) {
            _handleItemDataChange(notification);
            return true;
          },
        ),
      ),
    );
  }
}
