import 'package:spark_form/spark.dart';

enum VerifyType {
  /// verify value after changed
  change,

  /// verify value after blur
  blur,

  /// include change and blur
  all,
}

abstract class SparkValidator {
  SparkValidator({
    required this.errorMessage,
    required this.type,
    this.strategy,
  });

  /// verify error message
  String errorMessage;

  /// validate type
  VerifyType? type;

  /// validate strategy
  ///
  FormStrategy? strategy;

  /// validate rule
  ValidateResult validate(dynamic value, String fieldName, VerifyType? type);
}

class SparkLengthValidator extends SparkValidator {
  int? length;

  SparkLengthValidator.length(
    this.length, {
    required String errorMessage,
    required VerifyType? type,
    FormStrategy? strategy = FormStrategy.limit,
  }) : super(
          errorMessage: errorMessage,
          type: type,
          strategy: strategy,
        );

  @override
  ValidateResult validate(dynamic value, String fieldName, VerifyType? type) {
    return ValidateResult.from(
      errorMessage: errorMessage,
      fieldName: fieldName,
      success: value.length <= length,
      strategy: strategy,
      type: type,
    );
  }
}

class SparkFormatValidator extends SparkValidator {
  RegExp? rule;

  /// vilify result by regexp
  SparkFormatValidator.form(
    this.rule, {
    required String errorMessage,
    required VerifyType? type,
    FormStrategy? strategy = FormStrategy.limit,
  }) : super(
          errorMessage: errorMessage,
          type: type,
          strategy: strategy,
        );

  @override
  ValidateResult validate(dynamic value, String fieldName, VerifyType? type) {
    return ValidateResult.from(
      errorMessage: errorMessage,
      fieldName: fieldName,
      success: emptyOrNullString(value) || rule!.hasMatch(value),
      strategy: strategy,
      type: type,
    );
  }
}

class SparkNumberValidator extends SparkValidator {
  int integerLength = 500;

  int decimalLength = 500;

  double max = double.maxFinite;

  double min = double.minPositive;

  final rule = RegExp(r'^([0-9]{1,})+$|^([0-9]+\.)+$|^([0-9]+\.+[0-9]{1,})+$');

  /// number only
  SparkNumberValidator.number({
    required String errorMessage,
    required VerifyType? type,
    FormStrategy? strategy = FormStrategy.limit,
  }) : super(
          errorMessage: errorMessage,
          type: type,
          strategy: strategy,
        );

  /// setup number range
  SparkNumberValidator.range({
    required this.max,
    required this.min,
    required String errorMessage,
    required VerifyType? type,
    FormStrategy? strategy = FormStrategy.limit,
  })  : assert(
          max > min,
          'SparkNumberValidator.range max must be max than min',
        ),
        super(
          errorMessage: errorMessage,
          type: type,
          strategy: strategy,
        );

  /// setup number integer part length and decimal part length
  SparkNumberValidator.length({
    required this.integerLength,
    required this.decimalLength,
    required String errorMessage,
    required VerifyType? type,
    FormStrategy? strategy = FormStrategy.limit,
  }) : super(
          errorMessage: errorMessage,
          type: type,
          strategy: strategy,
        );

  @override
  ValidateResult validate(value, String fieldName, VerifyType? type) {
    bool _isSuccess = false;
    if (emptyOrNullString(value)) {
      _isSuccess = true;
    } else {
      bool _matchRule = rule.hasMatch(value);
      if (_matchRule) {
        List<String> _splitList = value.split('.');
        if (_splitList.length == 2) {
          if (integerLength >= _splitList.first.length &&
              decimalLength >= _splitList.last.length) {
            _isSuccess = true;
          }
        } else if (_splitList.length == 1) {
          if (integerLength >= _splitList.first.length) {
            _isSuccess = true;
          }
        }
        if (_isSuccess) {
          double? _number = double.tryParse(value);
          if (_number != null) {
            if (_number > max || _number < min) {
              _isSuccess = false;
            }
          }
        }
      }
    }
    return ValidateResult.from(
      errorMessage: errorMessage,
      fieldName: fieldName,
      success: _isSuccess,
      strategy: strategy,
      type: type,
    );
  }
}

class SparkRequiredValidator extends SparkValidator {
  SparkRequiredValidator.required({
    required String errorMessage,
    required VerifyType? type,
    FormStrategy? strategy = FormStrategy.limit,
  }) : super(
          errorMessage: errorMessage,
          type: type,
          strategy: strategy,
        );

  @override
  ValidateResult validate(value, String fieldName, VerifyType? type) {
    bool _result = value != null;
    if (value is List<SparkFormDataChangeImp>) {
      _result = value.any((element) => element.needSelected);
    }
    if (value is String) {
      _result = !emptyOrNullString(value);
    }
    return ValidateResult.from(
      errorMessage: errorMessage,
      fieldName: fieldName,
      success: _result,
      strategy: strategy,
      type: type,
    );
  }
}

class SparkComposeOrValidator extends SparkValidator {
  List<SparkValidator> validators;

  SparkComposeOrValidator.composeOr({
    required this.validators,
    required String errorMessage,
    required VerifyType? type,
    FormStrategy? strategy = FormStrategy.limit,
  }) : super(
          errorMessage: errorMessage,
          type: type,
          strategy: strategy,
        );

  @override
  ValidateResult validate(value, String fieldName, VerifyType? type) {
    Iterable<ValidateResult> _result =
        validators.map((_rule) => _rule.validate(value, fieldName, type));
    if (_result.any((element) => !element.success)) {
      return ValidateResult.from(
        success: false,
        fieldName: fieldName,
        errorMessage: errorMessage,
        type: type,
        strategy: strategy,
      );
    }
    return ValidateResult.from(success: true);
  }
}

class ValidateResult {
  ValidateResult.from({
    this.errorMessage,
    this.fieldName,
    required this.success,
    this.type,
    this.strategy,
  });

  /// result tip message
  String? errorMessage;

  /// result
  bool success;

  /// check type
  VerifyType? type;

  /// check field
  String? fieldName;

  /// use strategy type
  FormStrategy? strategy;
}
