part of spark_widget;

class SparkTagGroup extends SparkFormStateLessWidget<List<SparkTagGroupItem>> {
  const SparkTagGroup({
    Key? key,
    this.itemBuilder,
    required FormGroupItem manage,
    this.showClean = false,
    this.type = SparkTagGroupType.multiple,
    this.spacing = 10,
    this.runSpacing = 10,
    this.onAdd,
    this.addWidget,
    this.hitText,
    this.hitWidget,
    this.hitStyle,
    this.alignment,
    this.cleanIconBuilder,
    this.cleanSize,
  }) : super(key: key, manage: manage);

  /// custom item builder
  final TagGroupItemBuilder? itemBuilder;

  /// set up clean icon visible
  final bool showClean;

  /// group type
  final SparkTagGroupType type;

  /// see [Wrap]
  final double spacing;

  /// see [Wrap]
  final double runSpacing;

  /// group out side add function
  final SparkTagGroupItem? Function()? onAdd;

  /// custom add widget
  final Widget? addWidget;

  /// empty hit text
  final String? hitText;

  /// hit widget
  final Widget? hitWidget;

  /// hit text style
  final TextStyle? hitStyle;

  /// see [Wrap]
  final WrapAlignment? alignment;

  /// custom clean icon builder
  final TagGroupItemIconBuilder? cleanIconBuilder;

  final Size? cleanSize;

  void _handleTagDelete(BuildContext context, SparkTagGroupItem data) {
    fieldData.remove(data);
    handleFormDataChange(context, fieldData);
  }

  void _handleSelect(BuildContext context, SparkTagGroupItem data) {
    data.selected = !data.selected;
    if (type == SparkTagGroupType.singe ||
        type == SparkTagGroupType.singeAndIncrease) {
      for (SparkTagGroupItem _item in fieldData) {
        if (_item != data && _item.selected) {
          _item.selected = false;
        }
      }
    }
    handleFormDataChange(context, fieldData);
  }

  void _handleAddItem() {
    if (onAdd != null) {
      SparkTagGroupItem? _addItem = onAdd!();
      if (_addItem != null) {
        fieldData.add(_addItem);
        manage.update();
      }
    }
  }

  Widget _buildTag(
      BuildContext context, SparkTagGroupItem data, SparkTheme? theme) {
    if (itemBuilder != null) {
      return itemBuilder!(context, data);
    }
    return SparkTagGroupItemWidget(
      data: data,
      showTagIcon: showClean,
      onIconClick: () {
        _handleTagDelete(context, data);
      },
      isSelect: data.selected,
      decoration: theme?.tagItemDecoration,
      style: theme?.tagStyle,
      selectedDecoration: theme?.tagSelectedItemDecoration,
      padding: theme?.tagContentPadding,
      selectStyle: theme?.tagSelectedStyle,
      minWidth: theme?.tagMinWidth,
      iconBuilder: cleanIconBuilder,
    );
  }

  Widget? _buildAdd(BuildContext context, SparkTheme? theme) {
    if (type == SparkTagGroupType.multipleAndIncrease ||
        type == SparkTagGroupType.singeAndIncrease) {
      return GestureDetector(
        behavior: HitTestBehavior.opaque,
        onTap: () {
          _handleAddItem();
        },
        child: SparkTagGroupItemWidget(
          data: SparkTagGroupItem.from(name: ''),
          isSelect: false,
          decoration: theme?.tagAddDecoration,
          style: theme?.tagStyle,
          padding: theme?.tagContentPadding,
          minWidth: theme?.tagMinWidth,
          customWidget: addWidget ??
              const Icon(
                Icons.add,
                size: 14,
              ),
        ),
      );
    }
    return null;
  }

  Widget? _buildPlaceholder(BuildContext context, SparkTheme? theme) {
    if (fieldData.isNotEmpty) {
      assert(() {
        return !(hitText != null && hitWidget != null);
      }(), 'placeholder and placeholderWidget only set one');
      if (hitText != null) {
        return Text(
          hitText!,
          style: hitStyle ??
              TextStyle(
                fontSize: theme?.formItemFontSize,
                color: theme?.hitColor,
              ),
          maxLines: 1,
          overflow: TextOverflow.ellipsis,
        );
      }
      return hitWidget;
    }
    return null;
  }

  @override
  Widget build(BuildContext context) {
    SparkTheme? theme = SparkThemeData.of(context);
    Widget? _addWidget = _buildAdd(context, theme);
    Widget? _placeholderWidget = _buildPlaceholder(context, theme);
    return SparkWrapper(
      builder: (_ctx, _value, _updater) {
        return Wrap(
          spacing: spacing,
          alignment: alignment ?? WrapAlignment.start,
          runSpacing: runSpacing,
          children: [
            if (_placeholderWidget != null && fieldData.isEmpty)
              _placeholderWidget,
            for (SparkTagGroupItem _data in fieldData)
              GestureDetector(
                behavior: HitTestBehavior.opaque,
                onTap: () {
                  _handleSelect(context, _data);
                },
                child: _buildTag(context, _data, theme),
              ),
            if (_addWidget != null) _addWidget,
          ],
        );
      },
      manage: manage,
    );
  }
}

class SparkTagGroupItemWidget extends StatelessWidget {
  const SparkTagGroupItemWidget({
    Key? key,
    required this.data,
    this.maxLines = 1,
    this.maxWidth,
    this.padding,
    this.isSelect = false,
    this.selectedDecoration,
    this.decoration,
    this.style,
    this.selectStyle,
    this.showTagIcon = false,
    this.iconBuilder,
    this.onIconClick,
    this.minWidth,
    this.customWidget,
    this.cleanSize,
    this.offsetFactor = 0.35,
  }) : super(key: key);

  final SparkTagGroupItem data;

  final EdgeInsets? padding;

  final double? maxWidth;

  final double? minWidth;

  final int maxLines;

  final Decoration? selectedDecoration;

  final Decoration? decoration;

  final TextStyle? selectStyle;

  final TextStyle? style;

  final bool isSelect;

  final bool showTagIcon;

  final TagGroupItemIconBuilder? iconBuilder;

  final VoidCallback? onIconClick;

  final Widget? customWidget;

  final Size? cleanSize;

  /// offset factor, setup clean icon offset top and right side
  final double offsetFactor;

  @override
  Widget build(BuildContext context) {
    Size _cleanSize = cleanSize ?? const Size(14, 14);
    Offset _offset = Offset(
      _cleanSize.width * offsetFactor,
      _cleanSize.height * offsetFactor,
    );
    return Transform(
      transform: Matrix4.translationValues(-_offset.dx, -_offset.dy, 0),
      child: Stack(
        clipBehavior: Clip.none,
        children: [
          IntrinsicWidth(
            child: Container(
              padding: padding,
              margin: EdgeInsets.only(top: _offset.dx, right: _offset.dy),
              constraints: BoxConstraints(minWidth: minWidth ?? 0.0),
              width: maxWidth,
              alignment: Alignment.center,
              decoration: isSelect ? selectedDecoration : decoration,
              child: customWidget ??
                  Text(
                    data.name,
                    maxLines: maxLines,
                    style: isSelect ? selectStyle : style,
                  ),
            ),
          ),
          if (showTagIcon)
            Positioned(
              right: 0,
              top: 0,
              child: GestureDetector(
                onTap: onIconClick,
                behavior: HitTestBehavior.opaque,
                child: iconBuilder == null
                    ? UnconstrainedBox(
                        child: Image.asset(
                          'icons/icon_delete.png',
                          package: 'spark_form',
                          width: _cleanSize.width,
                          height: _cleanSize.height,
                        ),
                      )
                    : iconBuilder!(context, isSelect),
              ),
            ),
        ],
      ),
    );
  }
}

class SparkTagGroupItem extends SparkFormDataChangeImp {
  SparkTagGroupItem({
    required this.name,
    bool selected = false,
    dynamic data,
  })  : _selected = selected,
        super(data);

  SparkTagGroupItem.from({
    required this.name,
    bool selected = false,
    dynamic data,
  })  : _selected = selected,
        super(data);

  String name;

  bool _selected = false;

  bool get selected => _selected;

  set selected(bool value) {
    _selected = value;
    needSelected = value;
  }
}

enum SparkTagGroupType {
  /// pick multiple item
  multiple,

  /// pick one item
  singe,

  /// pick multiple and add item
  multipleAndIncrease,

  /// pick singe and add item
  singeAndIncrease,
}
